/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.amdatu.security.tokenprovider;

import java.util.SortedMap;

import javax.servlet.http.HttpServletRequest;

/**
 * This service facilitates generating tokens. A token is an encrypted set of attributes and a
 * signature. The signature is generated from these attributes and a private key, only known to
 * this token provider.
 * The common use case of creating and validating a token is as follows:
 * -1- First determine a set of attributes that the token must contain. This may include things like
 * a username, IP address, a timestamp, a nonce or a consumer key (for OAuth tokens).
 * -2- From this set of attributes a signature is created, using the generateSignature method. The signature
 * is creating by applying a hash algorithm onto the set of attributes plus a private key, only known by this
 * token provider.
 * -3- An unencrypted token is generated by converting the set of attributes to a single string and append it
 * to the signature of the attributes. An unencrypted token may for example be:
 * 'cebca7fb93e4c5316a6e91b3d7af470b username=Administrator ip=127.0.0.1'
 * -4- The token is encrypted using the AES encryption method, using the private key. This token is returned
 * by the generateToken method
 * -5- The encrypted token is added to the token store
 * -6- When a client wants to verify if a certain token is valid, it passes this encrypted token to this provider.
 * The provider first verifies that this token is available in the token store. Then it decrypts the token,
 * verifies that the signature is correct and then resolves the attributes stored in the token. It returns
 * those attributes.
 * -7- A token can be invalidated by removing it from the token store.
 * 
 */
public interface TokenProvider {
    /**
     * The PID of the configuration of this service.
     */
    String PID = "org.amdatu.security.tokenprovider";

    /**
     * Name of the cookie that stores an Amdatu token.
     */
    String TOKEN_COOKIE_NAME = "amdatu_token";

    /**
     * Name under which the nonce parameter is stored.
     */
    String NONCE = "token_nonce";

    /**
     * Name under which the timestamp parameter is stored.
     */
    String TIMESTAMP = "token_timestamp";

    /**
     * Name under which the tenant parameter is stored.
     */
    String TENANTID = "tenantid";

    /**
     * Name under which a username parameter can be stored (this is optional).
     */
    String USERNAME = "token_username";

    /**
     * Generates a new token for the specified set of token attributes. First of all a signature is created
     * from the set of attributes using a private key, only known by this token provider. The attributes are
     * converted to a single String and together with the signature this forms the unencrypted token.
     * Finally, the token is encrypted using aAES encryption method, again using a private key only known
     * by this token provider. This encrypted token is returned.
     * Note that to this list of attributes always a nonce and a timestamp are added. Therefore attributes
     * with names NONCE or TIMESTAMP are preserved and should not be used. A TokenProviderException is thrown
     * in case the map of attributes already contains these attributes.
     * 
     * @param attributes
     *        The attributes to create the token for. May be empty or null, in which case the
     *        token is generated from only a nonce and a timestamp.
     * @return The generated encrypted token
     * @throws TokenProviderException
     *         In case an unexpected error occurred during token generation.
     */
    String generateToken(SortedMap<String, String> attributes) throws TokenProviderException;

    /**
     * Verifies if the token is valid. If the token is valid, a map of attributes is returned which
     * are included in the token. This is the same list as provided when the token was generated, but
     * with additional nonce and timestamp attributes. If the token is invalid, a InvalidTokenException
     * is thrown.
     * 
     * @param encryptedToken
     *        The encrypted token to verify
     * @return List of original attributes, plus nonce and timestamp attributes.
     * @throws TokenProviderException
     *         In case an unexpected error occurred during token generation.
     * @throws InvalidTokenException
     *         In case the provided token is invalid
     */
    SortedMap<String, String> verifyToken(String encryptedToken) throws TokenProviderException, InvalidTokenException;

    /**
     * Invalidates the specified token. A token that has been invalidated cannot be used anymore.
     * 
     * @param encryptedToken
     *        the token to invalidate.
     */
    void invalidateToken(String encryptedToken);

    /**
     * Returns the request token associated with the specified request or null of none is associated.
     * 
     * @param request
     *        The request to get the token from
     * @return the request token associated with the specified request
     */
    String getTokenFromRequest(HttpServletRequest request);
}
